# Copyright 2021 Marcel Konstapel
# https://www.meezenest.nl/mees
#
# This file is part of SolarGeneratorDatalogger.
# A set of programs and scripts to log all the available data from a self build solar generator.
# It runs on a Raspberry Pi Zero 1 W with Raspbian
#
# It logs data from a EPEVER Soalrge Charge Controller (via Bluetooth) and an EPEVER solar charger (via RS485-USB).
# The data is analyzed and stored in a file. The data is also available via a web browser over Wifi
# (Raspberry Pi acts as access point).
# Future plan is to implement LoRa so the data can be accessed remotely.
#
# Program is configured via the file "config.json"
#
# Example of config_file:
#
#    {
#        "epever_data_file": "epever_data.json",
#        "RS485_port": "/dev/ttyXRUSB0",
#        "RS485_address": 1
#    }
#
# Usage of program:
#
#  -c, --config_file : reads the most important registers of the charge
#                      controller and writes it to the file specified in the
#                      config_file
#  -s, --set_date    : sets the date and time from the given argument with
#                      format: YYYY-MM-DD HH:MM:SS
#  -d, --dump_file   : dumps all the registers in the specified file
#
#
# A lot of effort is made to catch all possible communication errors, so the program never hangs.
# When an error occurs, the program exits and has probably not written any data to  'epever_data_file'
# But at least we did not crash! Check the date of the data file to make sure the data is written or not.
#
# Example of epever_data_file (JSON format):
#
#{
# "solar_voltage": 0,
# "solar_current": 0,
# "solar_power": 0,
# "load_voltage": 13.11,
# "load_current": 0,
# "load_power": 0,
# "battery_voltage": 13.14,
# "battery_current": 0,
# "battery_power": 0,
# "controller_temp": 18.1,
# "ambient_temp": 16.42,
# "generated_energy_today": 0,
# "date_time": "2021-11-12 15:56:52",
# "battery_status": {
#  "wrong_identifaction_for_rated_voltage": false,
#  "battery_inner_resistence_abnormal": false,
#  "temperature_warning_status": "NORMAL",
#  "battery_status": "NORMAL"
# },
# "controller_status": {
#  "input_voltage_status": "NORMAL",
#  "charging_mosfet_is_short_circuit": false,
#  "charging_or_anti_reverse_mosfet_is_open_circuit": false,
#  "anti_reverse_mosfet_is_short_circuit": false,
#  "input_over_current": false,
#  "load_over_current": false,
#  "load_short_circuit": false,
#  "load_mosfet_short_circuit": false,
#  "disequilibrium_in_three_circuits": false,
#  "pv_input_short_circuit": false,
#  "charging_status": "NO_CHARGING",
#  "fault": false,
#  "running": true
# }
#}
#
# This code is based on:
#   https://github.com/rosswarren/epevermodbus
#   https://minimalmodbus.readthedocs.io/en/master/index.html
#   https://github.com/kasbert/epsolar-tracer (Linux driver for CC-USB-RS485-150U USB to RS485 cable)
#
# Thanks for sharing your knowledge!
#
# V0.1 2021-11-12
#   - First working program
#   - Generic read works
#   - Dump all registers works
#   - TODO: Set date does not work yet
#   - TODO: Make LOAD-switch availible through command line arguments
#
# This Python3 program gathers the data from the Liontron BMS over Bluetooth BLE.
#
#    SolarGeneratorDatalogger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SolarGeneratorDatalogger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SolarGeneratorDatalogger.  If not, see <https://www.gnu.org/licenses/>.
#
import json
import sys
import time
import getopt

from json.decoder import JSONDecodeError
from datetime import datetime
from itertools import count
from multiprocessing import Process
from epever_control import EpeverChargeController

def dump_all_registers():

    rawdat={}
    #Address range 0x3000
    rawdat['rated_solar_voltage'] = controller.get_rated_solar_voltage()
    rawdat['rated_solar_current'] = controller.get_rated_solar_current()
    rawdat['rated_solar_power'] = controller.get_rated_solar_power_l() + 256*controller.get_rated_solar_power_h()
    rawdat['rated_charging_voltage'] = controller.get_rated_charging_voltage()
    rawdat['rated_charging_current'] = controller.get_rated_charging_current()
    rawdat['rated_charging_power'] = controller.get_rated_charging_power_l() + 256 * controller.get_rated_charging_power_h()
    rawdat['charging_mode'] = controller.get_charging_mode()
    rawdat['rated_load_current'] = controller.get_rated_load_current()
    # Address range 0x3100
    rawdat['solar_voltage'] = controller.get_solar_voltage()
    rawdat['solar_current'] = controller.get_solar_current()
    rawdat['solar_power'] = controller.get_solar_power_l() + 256 * controller.get_solar_power_h()
    rawdat['battery_voltage'] = controller.get_battery_voltage()
    rawdat['battery_current'] = controller.get_battery_current()
    rawdat['battery_power'] = controller.get_battery_power_l() + 256 * controller.get_battery_power_h()
    rawdat['load_voltage'] = controller.get_load_voltage()
    rawdat['load_current'] = controller.get_load_current()
    rawdat['load_power'] = controller.get_load_power_l() + 256 * controller.get_load_power_h()
    rawdat['battery_temperature'] = controller.get_battery_temperature()
    rawdat['controller_temperature'] = controller.get_controller_temperature()
    rawdat['heatsink_temperature'] = controller.get_heatsink_temperature()
    rawdat['battery_state_of_charge'] = controller.get_battery_state_of_charge()
    rawdat['remote_battery_temperature'] = controller.get_remote_battery_temperature()
    rawdat['battery_real_rated_voltage'] = controller.get_battery_real_rated_voltage()
    # Address range 0x3200
    rawdat['battery_status'] = controller.get_battery_status()
    rawdat['charging_equipment_status'] = controller.get_charging_equipment_status()
    # Address 0x3300
    rawdat['maximum_solar_voltage_today'] = controller.get_maximum_solar_voltage_today()
    rawdat['minimum_solar_voltage_today'] = controller.get_minimum_solar_voltage_today()
    rawdat['maximum_battery_voltage_today'] = controller.get_maximum_battery_voltage_today()
    rawdat['minimum_battery_voltage_today'] = controller.get_minimum_battery_voltage_today()
    rawdat['consumed_energy_today'] = controller.get_consumed_energy_today_l() + 256 * controller.get_consumed_energy_today_h()
    rawdat['consumed_energy_month'] = controller.get_consumed_energy_month_l() + 256 * controller.get_consumed_energy_month_h()
    rawdat['consumed_energy_year'] = controller.get_consumed_energy_year_l() + 256 * controller.get_consumed_energy_year_h()
    rawdat['consumed_energy_total'] = controller.get_consumed_energy_total_l() + 256 * controller.get_consumed_energy_total_h()
    rawdat['generated_energy_today'] = controller.get_generated_energy_today_l() + 256 * controller.get_generated_energy_today_h()
    rawdat['generated_energy_month'] = controller.get_generated_energy_month_l() + 256 * controller.get_generated_energy_month_h()
    rawdat['generated_energy_year'] = controller.get_generated_energy_year_l() + 256 * controller.get_generated_energy_year_h()
    rawdat['generated_energy_total'] = controller.get_generated_energy_total_l() + 256 * controller.get_generated_energy_total_h()
    rawdat['carbon_reduction'] = controller.get_carbon_reduction_l() + 256 * controller.get_carbon_reduction_h()
    rawdat['battery_net_current'] = controller.get_battery_net_current_l() + 256 * controller.get_battery_net_current_h()
    rawdat['battery_temperature'] = controller.get_battery_temperature()
    rawdat['ambient_temperature'] = controller.get_ambient_temperature()
    # Address range 0x9000
    rawdat['battery_type'] = controller.get_battery_type()
    rawdat['battery_capacity'] = controller.get_battery_capacity()
    rawdat['temperature_compensation_coefficient'] = controller.get_temperature_compensation_coefficient()
    rawdat['over_voltage_disconnect_voltage'] = controller.get_over_voltage_disconnect_voltage()
    rawdat['charging_limit_voltage'] = controller.get_charging_limit_voltage()
    rawdat['over_voltage_reconnect_voltage'] = controller.get_over_voltage_reconnect_voltage()
    rawdat['equalize_charging_voltage'] = controller.get_equalize_charging_voltage()
    rawdat['boost_charging_voltage'] = controller.get_boost_charging_voltage()
    rawdat['float_charging_voltage'] = controller.get_float_charging_voltage()
    rawdat['boost_reconnect_charging_voltage'] = controller.get_boost_reconnect_charging_voltage()
    rawdat['low_voltage_reconnect_voltage'] = controller.get_low_voltage_reconnect_voltage()
    rawdat['under_voltage_recover_voltage'] = controller.get_under_voltage_recover_voltage()
    rawdat['under_voltage_warning_voltage'] = controller.get_under_voltage_warning_voltage()
    rawdat['low_voltage_disconnect_voltage'] = controller.get_low_voltage_disconnect_voltage()
    rawdat['discharging_limit_voltage'] = controller.get_discharging_limit_voltage()
    rawdat['datetime'] = controller.get_datetime()
    rawdat['equalization_charging_cycle'] = controller.get_equalization_charging_cycle()
    rawdat['battery_temp_upper_warning_limit'] = controller.get_battery_temp_upper_warning_limit()
    rawdat['battery_temp_lower_warning_limit'] = controller.get_battery_temp_lower_warning_limit()
    rawdat['controller_inner_temp_upper_limit'] = controller.get_controller_inner_temp_upper_limit()
    rawdat['controller_inner_temp_upper_limit_recover'] = controller.get_controller_inner_temp_upper_limit_recover()
    rawdat['power_comp_temp_upper_limit'] = controller.get_power_comp_temp_upper_limit()
    rawdat['power_comp_temp_upper_limit_recover'] = controller.get_power_comp_temp_upper_limit_recover()
    rawdat['line_impedance'] = controller.get_line_impedance()
    rawdat['night_time_threshold_volt'] = controller.get_night_time_threshold_volt()
    rawdat['light_signal_startup_delay'] = controller.get_light_signal_startup_delay()
    rawdat['day_time_threshold_volt'] = controller.get_day_time_threshold_volt()
    rawdat['light_signal_turnoff_delay'] = controller.get_light_signal_turnoff_delay()
    rawdat['controling_mode'] = controller.get_controling_mode()
    rawdat['working_time_lenght_1'] = controller.get_working_time_lenght_1()
    rawdat['working_time_lenght_2'] = controller.get_working_time_lenght_2()
    rawdat['turn_on_time_1_sec'] = controller.get_turn_on_time_1_sec()
    rawdat['turn_on_time_1_min'] = controller.get_turn_on_time_1_min()
    rawdat['turn_on_time_1_hours'] = controller.get_turn_on_time_1_hours()
    rawdat['turn_off_time_1_sec'] = controller.get_turn_off_time_1_sec()
    rawdat['turn_off_time_1_min'] = controller.get_turn_off_time_1_min()
    rawdat['turn_off_time_1_hours'] = controller.get_turn_off_time_1_hours()
    rawdat['turn_on_time_2_sec'] = controller.get_turn_on_time_2_sec()
    rawdat['turn_on_time_2_min'] = controller.get_turn_on_time_2_min()
    rawdat['turn_on_time_2_hours'] = controller.get_turn_on_time_2_hours()
    rawdat['turn_off_time_2_sec'] = controller.get_turn_off_time_2_sec()
    rawdat['turn_off_time_2_min'] = controller.get_turn_off_time_2_min()
    rawdat['turn_off_time_2_hours'] = controller.get_turn_off_time_2_hours()
    rawdat['length_of_night'] = controller.get_length_of_night()
    rawdat['battery_rated_voltage'] = controller.get_battery_rated_voltage()
    rawdat['load_timing_control_selection'] = controller.get_load_timing_control_selection()
    rawdat['default_load_on_off_in_manual_mode'] = controller.get_default_load_on_off_in_manual_mode()
    rawdat['equalize_duration'] = controller.get_equalize_duration()
    rawdat['boost_duration'] = controller.get_boost_duration()
    rawdat['battery_discharge'] = controller.get_battery_discharge()
    rawdat['battery_charge'] = controller.get_battery_charge()
    rawdat['management_mode'] = controller.get_management_mode()
    # Address range 0x0
    rawdat['manual_control'] = controller.get_manual_control()
    rawdat['enable_load_test_mode'] = controller.get_enable_load_test_mode()
    rawdat['force_load_on'] = controller.get_force_load_on()
    # Address range 0x2000
    rawdat['is_device_over_temperature'] = controller.is_device_over_temperature()
    rawdat['is_night'] = controller.is_night()

    print (json.dumps(rawdat, indent=1, sort_keys=False))
    with open(dump_file, 'w') as f:
        json.dump(rawdat, f, indent=1) 
        

config_file =""
set_date =""
dump_file=""

argv = sys.argv[1:]
try:
    options, args = getopt.getopt(argv, "c:s:d:",
                               ["config_file =",
                                "set_date =",
                                "dump_file ="])
except:
    print("epever-charge-controller can read and write the registers of a EPEVER solar")
    print("charge controller using the Modbus protocol.")
    print("")
    print("-c, --config_file : reads the most important registers of the charge")
    print("                    controller and writes it to the file specified in the")
    print("                    config_file")
    print("-s, --set_date    : sets the date and time from the given argument with")
    print("                    format: YYYY-MM-DD HH:MM:SS")
    print("-d, --dump_file   : dumps all the registers in the specified file")
    sys.exit()
 
for name, value in options:
    if name in ['-c', '--config_file']:
        config_file = value
    elif name in ['-s', '--set_date']:
        set_date = value
    elif name in ['-d', '--dump']:
        dump_file = value
        
if not config_file:
    print("No configuration file given.")
    sys.exit();
    
# read configuration file
try:
    with open(config_file, 'r') as myfile:
        data=myfile.read()
except FileNotFoundError:
    print("Configuration file <"+ config_file +"> not found.")
    sys.exit();

# parse configuration file
try:
    obj = json.loads(data)
    tmp = obj['epever_data_file']
    tmp1 = obj['RS485_port']
    tmp2 = obj['RS485_address']
except JSONDecodeError as e:
    print("Error decoding configuration file. Not JSON format?")
    sys.exit()
except TypeError as e:
    print("Error decoding configuration file: JSON objects not str, bytes or bytearray.")
    sys.exit()
except KeyError:
    print("Error reading configuration file: needed fields not found.")
    sys.exit()

# show values from config file
print("Using RRS485 port: " + str(obj['RS485_port']))
print("Using RRS485 address: " + str(obj['RS485_address']))
print("Saving EPEVER data to: " + str(obj['epever_data_file']))

controller = EpeverChargeController(obj['RS485_port'], obj['RS485_address'])
    
if set_date:
    print("Setting the time and date.")
    datetime={}    
    datetime['year'] = 21
    datetime['month'] = 11
    datetime['day'] = 12
    datetime['hours'] = 20
    datetime['minutes'] = 41
    datetime['seconds'] = 0

    controller.write_date(datetime)


elif dump_file:
    print("Dumping all registers to file: "+ dump_file)
    dump_all_registers()
    
else:

    rawdat={}
    rawdat['solar_voltage'] = controller.get_solar_voltage()
    rawdat['solar_current'] = controller.get_solar_current()
    rawdat['solar_power'] = controller.get_solar_power_l() + 256*controller.get_solar_power_h()

    rawdat['load_voltage'] = controller.get_load_voltage()
    rawdat['load_current'] = controller.get_load_current()
    rawdat['load_power'] = controller.get_load_power_l() + 256*controller.get_load_power_h()

    rawdat['battery_voltage'] = controller.get_battery_voltage()
    rawdat['battery_current'] = controller.get_battery_current()
    rawdat['battery_power'] = controller.get_battery_power_l() + 256*controller.get_battery_power_h()

    rawdat['controller_temp'] = controller.get_controller_temperature()
    rawdat['ambient_temp'] = controller.get_ambient_temperature()

    rawdat['generated_energy_today'] = controller.get_generated_energy_today_l() + 256*controller.get_generated_energy_today_h()

    rawdat['date_time'] = controller.get_datetime()

    rawdat['battery_status'] = controller.get_battery_status()
    rawdat['controller_status'] = controller.get_charging_equipment_status()

    print (json.dumps(rawdat, indent=1, sort_keys=False))
    with open(obj['epever_data_file'], 'w') as f:
        json.dump(rawdat, f, indent=1) 

    #controller.switch_load_on()
